/* C99 */

#define _XOPEN_SOURCE
#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <Xm/LabelG.h>
#include <Xm/PushB.h>
#include <Xm/MessageB.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xmd/Icon.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include "icons/level0.xbm"
#include "icons/level1.xbm"
#include "icons/level2.xbm"
#include "icons/level3.xbm"
#include "icons/level4.xbm"
#include "icons/charging.xbm"
#include "icons/error.xbm"
#include "icons/not.xbm"

typedef enum {
	BatteryStateCharging,
	BatteryStateNotCharging,
	BatteryStateFull,
	BatteryStateDischarging,
	BatteryStateError
} BatteryState;

typedef struct {
	int          level;
	BatteryState state;
	int          hours;
	int          minutes;
	int          seconds;
} BatteryInfo;

BatteryInfo getBatteryInfo (void);
Pixmap      selectBatteryIcon (BatteryInfo);
void        loadAllPixmaps (Widget);
void        batteryPoll (XtPointer, XtIntervalId *);
void        resetBatteryPollTimeout(void);
void        batteryInfoDialog (Widget, XtPointer, XtPointer);

Pixmap levels[8] = { 0 };
XtAppContext application;
Widget icon;
Widget text;
Widget layout;
Widget window;

int main (int argc, char *argv[]) {
	window = XtVaAppInitialize (
		&application, "Battery",
		NULL, 0,
		&argc, argv,
		NULL,
		XmNtitle, "Battery",
		NULL);
	
	layout = XtVaCreateWidget (
		"layout", xmRowColumnWidgetClass, window,
		XmNorientation, XmHORIZONTAL,
		NULL);
	
	loadAllPixmaps(layout);
	icon = XtVaCreateManagedWidget (
		"batteryIcon", xmPushButtonWidgetClass, layout,
		XmNleftAttachment,  XmATTACH_FORM,
		XmNlabelType,       XmPIXMAP,
		NULL);
	XtAddCallback(icon, XmNactivateCallback, batteryInfoDialog, NULL);
	
	XtVaCreateManagedWidget (
		"separator", xmSeparatorWidgetClass, layout,
		XmNorientation, XmVERTICAL,
		NULL);
	
	text = XtVaCreateManagedWidget (
		"text", xmLabelGadgetClass, layout,
		XmNalignment,       XmALIGNMENT_CENTER,
		NULL);
		
	batteryPoll(NULL, NULL);

	XtManageChild(layout);
	XtRealizeWidget(window);
	XtAppMainLoop(application);
}

BatteryInfo getBatteryInfo (void) {
	BatteryInfo result = { 0 };
	char charging[16]  = { 0 };
	int battery;

	FILE *stream = popen("acpi -b", "r");
	if (stream == NULL) {
		result.state = BatteryStateError;
		return result;
	}
	fscanf (
		stream, "Battery %d: %16s %d%%, %d:%d:%d",
		&battery, charging, &result.level,
		&result.hours, &result.minutes, &result.seconds);
	
	switch (charging[0]) {
	case 'C': result.state = BatteryStateCharging;    break;
	case 'N': result.state = BatteryStateNotCharging; break;
	case 'D': result.state = BatteryStateDischarging; break;
	case 'F': result.state = BatteryStateFull;        break;
	default: result.state = BatteryStateError;        break;
	}
	if (pclose(stream) != 0) {
		result.state = BatteryStateError;
	}
	return result;
}

Pixmap selectBatteryIcon (BatteryInfo info) {
	if (info.state == BatteryStateCharging) {
		return levels[5];
	} else if (info.state == BatteryStateNotCharging) {
		return levels[7];
	} else if (info.state == BatteryStateError) {
		return levels[6];
	} else if (info.level < 15) {
		return levels[0];
	} else if (info.level < 35) {
		return levels[1];
	} else if (info.level < 65) {
		return levels[2];
	} else if (info.level < 85) {
		return levels[3];
	} else {
		return levels[4];
	}
}

void loadAllPixmaps (Widget widget) {
	levels[0] = XmdLoadBitmapIcon(widget, level0);
	levels[1] = XmdLoadBitmapIcon(widget, level1);
	levels[2] = XmdLoadBitmapIcon(widget, level2);
	levels[3] = XmdLoadBitmapIcon(widget, level3);
	levels[4] = XmdLoadBitmapIcon(widget, level4);
	levels[5] = XmdLoadBitmapIcon(widget, charging);
	levels[6] = XmdLoadBitmapIcon(widget, error);
	levels[7] = XmdLoadBitmapIcon(widget, not);
}

void batteryPoll (XtPointer clientData, XtIntervalId *timer) {
	(void)(clientData);
	(void)(timer);
	BatteryInfo info = getBatteryInfo();
	
	XtVaSetValues (
		icon,
		XmNlabelType,   XmPIXMAP,
		XmNlabelPixmap, selectBatteryIcon(info),
		NULL);
	XtVaSetValues (
		window,
		XmNiconPixmap, selectBatteryIcon(info),
		NULL);
	
	#define BUFFER_LEN 32
	char buffer[BUFFER_LEN];
	snprintf(buffer, BUFFER_LEN, "%d%%", info.level);
	#undef BUFFER_LEN
	
	/* if we don't unmanage and then re-manage the child, the text flies to
	 * the top which is rather annoying. this little quirk bent me over and
	 * fucked me for hours. */
	XtUnmanageChild(text);
	XmString string = XmStringCreateLocalized(buffer);
	XtVaSetValues (
		text,
		XmNlabelString, string,
		NULL);
	XmStringFree(string);
	XtManageChild(text);
	
	resetBatteryPollTimeout();
}

void resetBatteryPollTimeout (void) {
	XtAppAddTimeOut(application, 2000, batteryPoll, NULL);
}

void batteryInfoDialog (Widget button, XtPointer clientData, XtPointer callData) {
	(void)(clientData);
	(void)(callData);
	
	BatteryInfo info = getBatteryInfo();
	
	static const char *states[] = {
		"Charging",
		"Full",
		"Discharging",
		"Error"
	};
	
	#define BUFFER_LEN 48
	char timeBuffer[BUFFER_LEN] = { 0 };
	switch (info.state) {
	case BatteryStateCharging:
		snprintf (
			timeBuffer, BUFFER_LEN, "\n%d:%02d:%02d until charged.",
			info.hours, info.minutes, info.seconds);
		break;
	case BatteryStateDischarging:
		snprintf (
			timeBuffer, BUFFER_LEN, "\n%d:%02d:%02d until empty.",
			info.hours, info.minutes, info.seconds);
		break;
	default:
	}
	#undef BUFFER_LEN
	#define BUFFER_LEN 128
	char messageBuffer[BUFFER_LEN];
	snprintf (
		messageBuffer, BUFFER_LEN, "%d%%, %s%s",
		info.level, states[info.state], timeBuffer);
	#undef BUFFER_LEN

	Arg args[5] = { 0 };
	int n = 0;
	XmString message = XmStringCreateLocalized(messageBuffer);
	XtSetArg(args[n], XmNmessageString, message); n ++;
	XmString title = XmStringCreateLocalized("Battery Status");
	XtSetArg(args[n], XmNdialogTitle, title); n ++;
	
	Widget dialog = XmCreateInformationDialog(button, "batteryInfo", args, n);
	
	XmStringFree(message);
	XmStringFree(title);
	
	XtManageChild(dialog);
	XtPopup(XtParent(dialog), XtGrabNone);
}
